/*
  Copyright 2012-2013 by
  Centre de recherche INRIA Sophia-Antipolis Mediterranee, equipe APICS,
  Sophia Antipolis, France.

  Contributors S. Chevillard
  sylvain.chevillard@ens-lyon.org

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  It is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This program is an implementation of the Airy Ai function for positive values.

   Assume that res is a mpfr_t already initialized with internal precision prec,
   and assume that x is a mpfr_t containing a positive value.
   After the call to AiryAi(res, x) the internal precision of res may have
   changed and res contains a value such that |Ai(x)-res| < 2^(-prec-10)*Ai(x)
*/
#include <mpfr.h>
#include "Brown.h"


/* Evaluates simultaneously all necessary constants for the evaluation  */
/* of F and G auxiliary functions for the computation of AiryAi (the    */
/* notations refer to the article that comes with this implementation   */
/* -- see README file).                                                 */

/* Automatically generated with Sollya using implementconstant          */
/* and manually tuned afterward, in order to factorize computations.    */
/* To make Brown's constant available within Sollya:                    */
/* Compile Brown.c (with `make libBrown.a`, or alternatively:)
     gcc -fPIC -Wall -c Brown.c
     gcc -shared -o libBrown.a Brown.o -lgmp -lmpfr

   Then, within Sollya:
     mpfr_Browns_const = libraryconstant("./libBrown.a");
     midpointmode=on;

   For instance, the code for g0 is obtained with:
     implementconstant(sqrt(10*sqrt(10)*mpfr_Browns_const)/(40*pi));
 */

void mpfr_AiryAux_const(mpfr_ptr f0, mpfr_ptr f1, mpfr_ptr f2, mpfr_ptr g0, mp_prec_t precf, mp_prec_t precg) {
  /* Declarations */
  mpfr_t brown, pi, pik, sqrt10, tmp, tmp2;

  /* Initializations */
  mpfr_init2 (brown, MAX(precf+6, precg+10));
  mpfr_init2 (pi, MAX(precf+16, precg+4));
  mpfr_init2 (pik, precf+13);
  mpfr_init2 (sqrt10, MAX(precf+8, precg+12));
  mpfr_init2 (tmp, precf+6);
  mpfr_init2 (tmp2, precg+10);

  /* Precomputations */
  mpfr_Browns_const (brown, precf+6);
  mpfr_const_pi (pi, GMP_RNDN);
  mpfr_sqrt_ui (sqrt10, 10, GMP_RNDN);

  /* f1 = [sqrt(3)/(6*pi)]; */
  mpfr_sqrt_ui (tmp, 3, GMP_RNDN);
  mpfr_set_prec (f1, precf+4);
  mpfr_mul_ui (f1, pi, 6, GMP_RNDN);
  mpfr_div (f1, tmp, f1, GMP_RNDN);

  /* g0 = [sqrt(10*sqrt(10)*mpfr_Browns_const)/(40*pi)] */
  /*  i.e. 1/(9*Gamma(2/3)^3)                           */
  mpfr_mul (tmp2, brown, sqrt10, GMP_RNDN);
  mpfr_mul_ui (tmp2, tmp2, 10, GMP_RNDN);
  mpfr_sqrt (tmp2, tmp2, GMP_RNDN);
  mpfr_set_prec (g0, precg+4);
  mpfr_mul_ui (g0, pi, 40, GMP_RNDN);
  mpfr_div (g0, tmp2, g0, GMP_RNDN);


  /* f0 = [(mpfr_Browns_const/(16*Pi^2*sqrt(10)))^(1/3)]; */
  /*  i.e. 3^(2/3) / ( 9*Gamma(2/3)^2 )                   */
  mpfr_sqr (pik, pi, GMP_RNDN); /* pik ~= pi^2 */
  mpfr_mul (tmp, sqrt10, pik, GMP_RNDN);
  mpfr_mul_ui (tmp, tmp, 16, GMP_RNDN);
  mpfr_div (tmp, brown, tmp, GMP_RNDN);
  mpfr_set_prec (f0, precf+2);
  mpfr_cbrt (f0, tmp, GMP_RNDN);

  /* f2 = [(sqrt(10)/(108 * mpfr_Browns_const * (pi^2)^2))^(1/3)]; */
  /*  i.e. 3^(1/3)*Gamma(2/3)^2 / (4pi^2)                          */
  mpfr_sqr (pik, pik, GMP_RNDN); /* pik ~= pi^4 */
  mpfr_mul (tmp, pik, brown, GMP_RNDN);
  mpfr_mul_ui (tmp, tmp, 108, GMP_RNDN);
  mpfr_div (tmp, sqrt10, tmp, GMP_RNDN);
  mpfr_set_prec (f2, precf+2);
  mpfr_cbrt (f2, tmp, GMP_RNDN);

  /* Cleaning stuff */
  mpfr_clear(brown);
  mpfr_clear(pi);
  mpfr_clear(pik);
  mpfr_clear(sqrt10);
  mpfr_clear(tmp);
  mpfr_clear(tmp2);
}


/* Evaluation by a straightforward algorithm */
/* Code for F automatically generated by metaMPFR. */
/* and manually simplified, so that we do not care about special cases or correct rounding. */
int
AiryAi (mpfr_ptr res, mpfr_srcptr x)
{
  /* Variables for the evaluation of F */
  mpfr_prec_t wprec;             /* working precision */
  mpfr_prec_t prec;              /* target precision for the whole evaluation */
  mpfr_prec_t precf;             /* target precision for the evaluation of F */
  mpfr_prec_t precg;             /* target precision for the evaluation of G */
  long int assumed_expF;
  unsigned long int k;
  int i;
  mpfr_t sf;                     /* used to store the partial sum of F */
  mpfr_t x3;                     /* used to store 2*x^3 */
  mpfr_t tmp;
  mpfr_t x1, x2;                 /* used to store x^i */
  mpfr_t tip0, tip1, tip2; /* used to store successive values of t_i */
  mpfr_t tmp_sp;
  int test0, test1, test2;
  int global_test;               /* used to test when the sum can be stopped */

  /* Variables for the evaluation of G */
  mpfr_t a, b, z, g0;
  mpfr_t v1, v2, sg, tmpg; /* Note also that x3 will be reused for storing x^3 */
  unsigned long int N, Nmin, Nmax, guard_bits, N0;
  long int E;


  /* In the end, we ensure a total relative error smaller than 2^(1-prec) */
  prec = MPFR_PREC (res) + 11;
  precf = prec + 3;
  precg = prec + 4;




  /********************************************************************/
  /***                    Precomputations for G                     ***/
  /********************************************************************/
  mpfr_init2(a, MPFR_SMALL_PRECISION);
  mpfr_init2(b, MPFR_SMALL_PRECISION);
  mpfr_init2(z, MPFR_SMALL_PRECISION);

  mpfr_set_d(a, 0.30000000004656613, GMP_RNDU); /* exact after conversion to double */
  mpfr_sqr(z, x, GMP_RNDU);
  mpfr_mul(z, z, x, GMP_RNDU); /* z >~ x^3 */
  mpfr_mul(a, a, z, GMP_RNDU);
  mpfr_sqrt(a, a, GMP_RNDU); /* a >~ sqrt( (3/10)*x^3 ) */
  N = mpfr_get_ui(a, GMP_RNDU);
  if (N >= 2) N = N-1; else N = 1;

  if (mpfr_get_exp(x) <= -1) { /* if x < 1/2 */
    E = precg + 6;
    mpfr_sqrt(z, z, GMP_RNDU); /* z >~ x^(3/2) */

    /* This is indeed for the computation of F */
    /* Set assumed_expF, so that F(x) >= 2^(-assumed_expF) */
    /* Uses F(x) >= 2^(-3)(1+x/2), valid for x<0.5 */
    mpfr_div_2ui(a, x, 1, GMP_RNDD);
    mpfr_add_ui(a, a, 1, GMP_RNDD);
    mpfr_div_2ui(a, a, 3, GMP_RNDD);
    assumed_expF = 1-mpfr_get_exp(x);
  }
  else {
    mpfr_sqrt(z, z, GMP_RNDU); /* z >~ x^(3/2) */
    /* The following constant is <~ (2/3)*log2(e)*(1-gamma6),
       where gamma6 = 6u/(1-6u) with u=2^(1-MPFR_SMALL_PRECISION)=2^(1-32)
       This ensures that a*z <~ (2/3)*log2(e)*x^(3/2) */
    mpfr_set_d(a, 0.9617966911755502, GMP_RNDD); /* exact after conversion to double */
    mpfr_mul(a, a, z, GMP_RNDD);
    E = 3*mpfr_get_exp(x);
    E = (E%4==0)?(E/4):(1+(E/4));
    E = precg + 8 + E - mpfr_get_si(a, GMP_RNDD);

    /* This is indeed for the computation of F */
    /* Set assumed_expF, so that F(x) >= 2^(-assumed_expF) */
    /* Uses F(x) >= (1/32)*x^(-1/2)*exp( (4/3)*x^(3/2) ), valid for x>0.5 */
    assumed_expF = mpfr_get_exp(x);
    assumed_expF = (assumed_expF%2==0)?(assumed_expF/2):(1+(assumed_expF/2));
    assumed_expF = assumed_expF + 5 -(mpfr_get_si(a, GMP_RNDD) + 1);
  }

  mpfr_set_d(a, 1.1036383234895766, GMP_RNDD); /* exact after conversion to double */
  mpfr_div(b, a, z, GMP_RNDD); /* b <~ 3/(ex^(3/2)) */
  mpfr_mul_ui(a, b, N, GMP_RNDD);
  mpfr_pow_ui(a, a, 2*N, GMP_RNDD); /* a <~ (9*N^2 / (e^2 * x^3))^N */
  while (mpfr_get_exp(a) < E+1) {
    mpfr_sqr(a, a, GMP_RNDD);
    mpfr_mul_2ui(a, a, 4*N, GMP_RNDD);
    N = N*2;
  }

  Nmin = N/2; /* Nmin does not achieve the accuracy */
  Nmax = N; /* Nmax does */

  while (Nmin+1 != Nmax) {
    N = (Nmin + Nmax)/2;
    mpfr_mul_ui(a, b, N, GMP_RNDD);
    mpfr_pow_ui(a, a, 2*N, GMP_RNDD);
    if (mpfr_get_exp(a)<E+1) Nmin = N;
    else Nmax = N;
  }

  N = Nmax;

  /* We now choose the truncation rank for beginning Miller's algorithm */
  if (mpfr_get_exp(x) <= -1) { /* if x < 1/2 */
    mpfr_set_d(b, -0.15625, GMP_RNDD); /* exact */
    mpfr_mul(a, b, x, GMP_RNDD);
    mpfr_add_ui(a, a, 1, GMP_RNDD); /* a <~ 1 - 0.15625x */
    mpfr_set_ui(b, precg+1-mpfr_get_exp(a), GMP_RNDU);
  }
  else {
    mpfr_set_d(a, 1.5215516914613545, GMP_RNDU); /* exact after conversion to double */
    mpfr_mul(b, a, z, GMP_RNDU);
    mpfr_add_ui(b, b, precg+2, GMP_RNDU); /* b >~ precg + 2 +
                                                      (2/3)*log2(e)*(sqrt(20/3)-1)*x^(3/2) */
  }
  mpfr_set_d(a, 0.3653681296855211, GMP_RNDU); /* a >~ 1/log2(20/3) */
  mpfr_mul(a, a, b, GMP_RNDU);
  N0 = mpfr_get_ui(a, GMP_RNDU);
  if (N0<N) N0 = N;
  if (N0<2) N0=2;


  /* We now find the guard_bits to add to t_target in order to obtain t_work */
  guard_bits = 0;
  Nmax = 1; /* Nmax = 2^(guard_bits) */
  while (Nmax < 128*(N+3)) { Nmax = Nmax*2; guard_bits++; }

  i=0;
  Nmax=1; /* Nmax = 2^i */
  while (Nmax < N0+2) { Nmax = Nmax*2; i++; }
  /* Now, i = ceil( log2(R+2) ) */
  if (i>(signed)precg+(signed)guard_bits-9) guard_bits = i+9-(signed)precg;


  mpfr_init2(g0, precg+guard_bits+4);


  /********************************************************************/
  /***                       Evaluation of F                        ***/
  /********************************************************************/

  /* We assume that the truncation rank will be smaller than precf */
  wprec = precf + MPFR_INT_CEIL_LOG2 ((unsigned)precf) + 6;
  mpfr_init (x3);
  mpfr_init (tmp);
  mpfr_init (x1);
  mpfr_init (x2);
  mpfr_init (tip0);
  mpfr_init (tip1);
  mpfr_init (tip2);
  mpfr_init (sf);
  mpfr_init2 (tmp_sp, MPFR_SMALL_PRECISION);

  /* Loop to ensure final relative error (should be useless in this case since assumed_exponent should always be correct */
  for (i=0;i<=1;i++)
    {
      mpfr_set_prec (x3, wprec);
      mpfr_set_prec (tmp, wprec);
      if(mpfr_get_prec(x) > wprec)
        mpfr_set_prec (x1, wprec);
      else
        mpfr_set_prec(x1, mpfr_get_prec(x));
      mpfr_set_prec (x2, wprec);
      mpfr_set_prec (tip0, wprec);
      mpfr_set_prec (tip1, wprec);
      mpfr_set_prec (tip2, wprec);
      mpfr_set_prec (sf, wprec);

      mpfr_set_ui (sf, 0, GMP_RNDN);
      mpfr_set (x1, x, GMP_RNDN);
      mpfr_mul (x2, x1, x1, GMP_RNDN);
      mpfr_mul (x3, x2, x1, GMP_RNDN);

      mpfr_AiryAux_const(tip0, tip1, tip2, g0, wprec, precg+guard_bits);

      mpfr_add (sf, sf, tip0, GMP_RNDN);
      mpfr_mul (tip1, tip1, x1, GMP_RNDN);
      mpfr_add (sf, sf, tip1, GMP_RNDN);
      mpfr_mul (tip2, tip2, x2, GMP_RNDN);
      mpfr_add (sf, sf, tip2, GMP_RNDN);

      mpfr_mul_si (x3, x3, 2, GMP_RNDN);

      /* Evaluation of the series */
      k = 3;
      for (;;)
        {
          mpfr_mul (tip0, tip0, x3, GMP_RNDN);
          mpfr_mul_si (tip0, tip0, 2*k-5, GMP_RNDN);
          mpfr_div_si3 (tip0, tip0, k, k-2, k-1, GMP_RNDN);
          mpfr_set (tmp, tip0, GMP_RNDN);
          mpfr_add (sf, sf, tmp, GMP_RNDN);

          k += 1;
          mpfr_mul (tip1, tip1, x3, GMP_RNDN);
          mpfr_mul_si (tip1, tip1, 2*k-5, GMP_RNDN);
          mpfr_div_si3 (tip1, tip1, k, k-2, k-1, GMP_RNDN);
          mpfr_set (tmp, tip1, GMP_RNDN);
          mpfr_add (sf, sf, tmp, GMP_RNDN);

          k += 1;
          mpfr_mul (tip2, tip2, x3, GMP_RNDN);
          mpfr_mul_si (tip2, tip2, 2*k-5, GMP_RNDN);
          mpfr_div_si3 (tip2, tip2, k, k-2, k-1, GMP_RNDN);
          mpfr_set (tmp, tip2, GMP_RNDN);
          mpfr_add (sf, sf, tmp, GMP_RNDN);

          k += 1;
          /* Checking that 2*(2*k-5)*x^3/(k*(k-1)*(k-2)) <= 1/2 */
          mpfr_set (tmp_sp, x, GMP_RNDU);
          mpfr_pow_ui (tmp_sp, tmp_sp, 3, GMP_RNDU);
          mpfr_mul_si (tmp_sp, tmp_sp, 2*k-5, GMP_RNDU);
          mpfr_div_si3 (tmp_sp, tmp_sp, k, k-2, k-1, GMP_RNDU);
          global_test = (MPFR_EXP(tmp_sp) <= -1);

          test0 = ( (!MPFR_IS_ZERO(sf))
                    && ( MPFR_IS_ZERO(tip0)
                         || (MPFR_EXP(tip0) + (mp_exp_t)precf + 4 <= MPFR_EXP(sf))
                         )
                    );
          test1 = ( (!MPFR_IS_ZERO(sf))
                    && ( MPFR_IS_ZERO(tip1)
                         || (MPFR_EXP(tip1) + (mp_exp_t)precf + 4 <= MPFR_EXP(sf))
                         )
                    );
          test2 = ( (!MPFR_IS_ZERO(sf))
                    && ( MPFR_IS_ZERO(tip2)
                         || (MPFR_EXP(tip2) + (mp_exp_t)precf + 4 <= MPFR_EXP(sf))
                         )
                    );
          if (test0 && test1 && test2 && global_test)
            break;
        }

      /* We post-check our assumption */
      if ( k < (unsigned)precf ) break;
      else wprec = precf + (MPFR_INT_CEIL_LOG2(k) + 1) + 6;
    } /* End of loop for ensuring final relative error  */


  /********************************************************************/
  /***                       Evaluation of G                        ***/
  /********************************************************************/
  /* Miller's algorithm */
  mpfr_init2(v1, precg+guard_bits);
  mpfr_init2(v2, precg+guard_bits);
  mpfr_init2(tmpg, precg+guard_bits);
  mpfr_init2(sg, precg+guard_bits);

  mpfr_set_prec(x3, precg+guard_bits);

  mpfr_set(tmpg, x, GMP_RNDN);
  mpfr_sqr(x3, tmpg, GMP_RNDN);
  mpfr_mul(x3, x3, tmpg, GMP_RNDN);
  mpfr_set_ui(v1, 1, GMP_RNDN); mpfr_set_ui(v2, 0, GMP_RNDN); /* (v1 v2) = v_N0 */
  while(N0 != 0) {
    N0 = N0-1;
    /* v_N computed from v_(N+1) */
    mpfr_set(tmpg, v1, GMP_RNDN); /* exact */
    mpfr_mul_ui2(v2, v2, 3*N0+4, 3*N0+5, GMP_RNDN);
    mpfr_div_ui2(v2, v2, N0+1, N0+2, GMP_RNDN);
    mpfr_mul_ui(v1, v1, 10, GMP_RNDN);
    mpfr_sub(v1, v1, v2, GMP_RNDN);
    mpfr_set(v2, tmpg, GMP_RNDN); /* exact */

    if (N0 == N-1) mpfr_set(sg, v1, GMP_RNDN);
    if (N0 < N-1) {
      mpfr_mul(sg, sg, x3, GMP_RNDN);
      mpfr_div_ui2(sg, sg, N0+1, N0+1, GMP_RNDN);
      mpfr_add(sg, sg, v1, GMP_RNDN);
    }
  }
  mpfr_div(sg, sg, v1, GMP_RNDN);
  mpfr_mul(sg, sg, g0, GMP_RNDN);




  /********************************************************************/
  /***                     Final reconstruction                     ***/
  /********************************************************************/
  /* We have
       precf = prec + 3
       precg = prec + 4
     We have |sg-G(x)| <= G(x)*3*2^(-precg) <= G(x)*2^(1-precf)
             |sf-F(x)| <= F(x)*2^(1-precf)
     Hence sg/sf = G(x)/F(x) <3> at precision precf
     So |sg/sf-G(x)/F(x)| <= 6*(G(x)/F(x))*2^(1-precf) <= (G(x)/F(x))*2^(1-prec)
  */

  mpfr_set_prec(res, precf);
  mpfr_div(res, sg, sf, GMP_RNDN);


  /* Clean up */
  mpfr_clear (tmp_sp);
  mpfr_clear (sf);
  mpfr_clear (tmp);
  mpfr_clear (x1);
  mpfr_clear (x2);
  mpfr_clear (tip0);
  mpfr_clear (tip1);
  mpfr_clear (tip2);

  mpfr_clear(z);
  mpfr_clear(a);
  mpfr_clear(b);

  mpfr_clear(v1);
  mpfr_clear(v2);
  mpfr_clear(tmpg);
  mpfr_clear(sg);
  mpfr_clear(x3);
  mpfr_clear(g0);

  return 0;
}
