/*
  Copyright 2012-2013 by
  Centre de recherche INRIA Sophia-Antipolis Mediterranee, equipe APICS,
  Sophia Antipolis, France.

  Contributors S. Chevillard
  sylvain.chevillard@ens-lyon.org

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  It is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <mpfr.h>
#include "AiryAi.h"


#define NUMBER_OF_CHECKS 10000
#define MAXIMAL_PREC 10000.
#define MAX_X 20

int initialization(gmp_randstate_t state) {
  mpz_t zseed;
  const char seed[] = "KOjFIs3JkN52qxcmnFXShLWnnKqF9o6VMEEo17ZDKzxvDm5at6Rsejxn3OlQ3rBZ1t9ZGU14HNf1KD19F60Tbpe5o0MXKQKn";

  gmp_randinit_default(state);
  mpz_init(zseed);
  mpz_set_str (zseed, seed, 62);
  gmp_randseed(state, zseed);
  srand(mpz_get_ui(zseed));
  mpz_clear(zseed);

  return 0;
}

/* Tests the implementation of Ai.                    */
/* We call AiryAi with several values of x>0          */
/* and several precisions prec.                       */
/* Each time, we check by comparing with mpfr_ai      */
/* computed with 50 guard bits. If the test fails,    */
/* it does not necessarily mean that AiryAi is        */
/* wrong (because of possible double rounding: our    */
/* implementation could give an error smaller to      */
/* 2^(-prec-10) with respect to the true value Ai(x)  */
/* and simultaneously give an error slightly larger   */
/* with respect to the value computed by MPFR with    */
/* 50 guard bits. This is very unlikely though).      */
int main(void) {
  int i;
  int test=1;
  mp_prec_t prec;
  mpfr_t x,yref, epsilon;
  mpfr_t y;
  gmp_randstate_t state;

  initialization(state);

  mpfr_init(x);
  mpfr_init(yref);
  mpfr_init(epsilon);

  for(i=0; i< NUMBER_OF_CHECKS; i++) {
    /* prec uniformly distributed in [MPFR_PREC_MIN, MAXIMAL_PREC+MPFR_PREC_MIN] */
    prec = MPFR_PREC_MIN+(mp_prec_t)(MAXIMAL_PREC*(double)rand()/(double)(RAND_MAX));
    mpfr_set_prec(x, prec);
    do {
      mpfr_urandomb (x, state); /* x uniformly distributed in (0,1] */
    } while (mpfr_zero_p(x));
    mpfr_mul_ui(x, x, MAX_X, GMP_RNDN); /* x uniformly distributed in (0, MAX_X] */

    prec = MPFR_PREC_MIN+(mp_prec_t)(MAXIMAL_PREC*(double)rand()/(double)(RAND_MAX));
    mpfr_set_prec(yref, prec+50);
    mpfr_set_prec(epsilon, prec+50);
    mpfr_ai(yref, x, GMP_RNDN);

    mpfr_init2(y, prec);
    AiryAi(y, x);

    mpfr_sub(epsilon, y, yref, GMP_RNDN);
    mpfr_div(epsilon, epsilon, yref, GMP_RNDN);

    if (mpfr_get_exp(epsilon) > -(signed)prec-10) { /* The implementation is supposed to provide an accuracy
                                                       of prec+10 bits */
      test = 0;
      printf("Test %d failed:\n", i);
      printf("Precision of x: %lu\n", (unsigned long)mpfr_get_prec(x));
      printf("x = "); mpfr_out_str(stdout, 2, 0, x, GMP_RNDN); printf("\n");
      printf("x ~ "); mpfr_out_str(stdout, 10, 10, x, GMP_RNDN); printf("\n");
      printf("Required precision: %lu\n", (unsigned long)prec);
      printf("Working precision: %lu\n", (unsigned long)mpfr_get_prec(y));
      printf("y      =     "); mpfr_out_str(stdout, 2, 0, y, GMP_RNDN); printf("\n");
      printf("True value = "); mpfr_out_str(stdout, 2, 0, yref, GMP_RNDN); printf("\n");
      printf("y      ~     "); mpfr_out_str(stdout, 10, 10, y, GMP_RNDN); printf("\n");
      printf("True value = "); mpfr_out_str(stdout, 10, 10, yref, GMP_RNDN); printf("\n");
      printf("Actual accuracy 2^{%ld} (should be smaller than 2^{%ld})\n\n", mpfr_get_exp(epsilon), -(long)prec);
    }

    mpfr_clear(y);
  }

  if (test) printf("All tests succeed!\n");
  gmp_randclear(state);
  mpfr_clear(x);
  mpfr_clear(epsilon);
  mpfr_clear(yref);
  mpfr_free_cache();
  return 0;
}
