/*
  Copyright 2012-2013 by
  Centre de recherche INRIA Sophia-Antipolis Mediterranee, equipe APICS,
  Sophia Antipolis, France.

  Contributors S. Chevillard
  sylvain.chevillard@ens-lyon.org

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  It is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <mpfr.h>
#include "timer.h"
#include "AiryAi.h"


/* This program compares our new implementation AiryAi with the          */
/* implementation mpfr_ai available with MPFR, when used to evaluate     */
/* Ai(x) at precision t, for x and t within two ranges. For each         */
/* couple (x,t), the best method is determined and a line of the form    */
/*                "x   t   mpfr   our    method"                         */
/* is written in file tune.dat, where 'mpfr' is the time (in             */
/*  microseconds spent in the evaluation of Ai(x) at precision t by MPFR */
/* and 'our' is the time spent by our algorithm.                         */
/* method \in {1,2} indicates the best method (1=mpfr, 2=our).           */


/* x will be chosen logarithmically distributed between 2^{MIN_EXPONENT} */
/* and 2^{MAX_EXPONENT+1} with POINTS_X sample points.                   */
/* t will be  chosen logarithmically distributed between MIN_PRECISION   */
/* and MAX_PRECISION with POINTS_PREC sample points.                     */

/* In order to determine the time spent in one execution of the          */
/* function f, we successively run f, during at least                    */
/* TIME_OF_A_SINGLE_SAMPLE microseconds. Then we divide the total time   */
/* by the number of times that f has been executed.                      */

#define MIN_EXPONENT 0
#define MAX_EXPONENT 9
#define MIN_PRECISION 10
#define MAX_PRECISION 2000
#define POINTS_PREC 450
#define POINTS_X 450
#define TIME_OF_A_SINGLE_SAMPLE 1000.

int initialization(gmp_randstate_t state) {
  mpz_t zseed;
  const char seed[] = "1BLg5Vq84iukX4SPS2MnBo1dhAwFd8FBWD3kztdhpWsHy3zTUluAsG1y7xOgDwOIFKMSIdKg4eM6yxgUKxpPp51iKnZj2bQAB";

  gmp_randinit_default(state);
  mpz_init(zseed);
  mpz_set_str (zseed, seed, 62);
  gmp_randseed(state, zseed);
  srand(mpz_get_ui(zseed));
  mpz_clear(zseed);

  return 0;
}

/* Returns the average time needed to evalute Ai(x) with a relative error */
/* smaller than 2^{-prec}, by mpfr_ai, expressed in microseconds.         */
double timing_mpfr(mpfr_t x, mp_prec_t prec) {
  timer_type timer;
  mpfr_t y;
  int i, count;
  double res;

  timer_init(timer); mpfr_init2(y, prec); count = 0;

  timer_start(timer); mpfr_ai(y, x, GMP_RNDN); timer_stop(timer);

  if (timer_get(timer) < TIME_OF_A_SINGLE_SAMPLE) {
    count = (int)(TIME_OF_A_SINGLE_SAMPLE/timer_get(timer));
    if (count==1) count=2;
    timer_init(timer);
    timer_start(timer);
    for(i=0;i<count;i++) mpfr_ai(y, x, GMP_RNDN);
    timer_stop(timer);
    res = timer_get(timer)/(double)count;
  }
  else res = timer_get(timer);
  mpfr_clear(y);

  return res;
}

/* Returns the average time needed to evalute Ai(x) with a relative error */
/* smaller than 2^{-prec}, by our implementation.                         */
double timing_new(mpfr_t x, mp_prec_t prec) {
  timer_type timer;
  mpfr_t y;
  int i, count;
  double res;

  timer_init(timer); mpfr_init2(y, prec); count = 0;

  timer_start(timer); AiryAi(y, x); timer_stop(timer);

  if (timer_get(timer) < TIME_OF_A_SINGLE_SAMPLE) {
    count = (int)(TIME_OF_A_SINGLE_SAMPLE/timer_get(timer));
    if (count==1) count=2;
    timer_init(timer);
    timer_start(timer);
    for(i=0;i<count;i++) {mpfr_set_prec(y, prec); mpfr_ai(y, x, GMP_RNDN);}
    timer_stop(timer);
    res = timer_get(timer)/(double)count;
  }
  else res = timer_get(timer);
  mpfr_clear(y);

  return res;
}


int main(void) {
  double incr_prec;
  mpfr_t incr_x;
  double eq1, eq2;
  mpfr_t temp, temp2;
  mpfr_t x0, x;
  mp_prec_t prec;
  gmp_randstate_t state;
  mpz_t mant;
  int i,j, best;
  FILE *output;

  initialization(state);

  output = fopen("tune.dat", "w");
  if (output == NULL) {
    printf("Cannot open tune.dat for writing.\n");
    exit(1);
  }

  /* Chooses x0 randomly with exponent MIN_EXPONENT */
  mpfr_init2(x0, MAX_PRECISION);
  mpz_init(mant);
  mpz_urandomb(mant, state, MAX_PRECISION);
  mpfr_set_z(x0, mant, GMP_RNDN);
  x0->_mpfr_sign = 1;
  mpfr_set_exp(x0, MIN_EXPONENT);
  mpz_clear(mant);
  gmp_randclear(state);

  /* Chooses incr_prec such that, starting from MIN_PRECISION */
  /* and multiplying POINTS_PREC times by incr_prec, we get   */
  /* something approximately equal to MAX_PRECISION.          */
  mpfr_init2(temp, 30);
  mpfr_init2(temp2, 30);
  mpfr_set_ui(temp, MAX_PRECISION, GMP_RNDU);
  mpfr_div_ui(temp, temp, MIN_PRECISION, GMP_RNDU);
  mpfr_set_ui(temp2, POINTS_PREC, GMP_RNDD);
  mpfr_ui_div(temp2, 1, temp2, GMP_RNDU);
  mpfr_pow(temp, temp, temp2, GMP_RNDU);
  incr_prec = mpfr_get_d(temp, GMP_RNDU);

  /* Chooses incr_x such that, starting from 2^MIN_EXPONENT   */
  /* and multiplying POINTS_X times by incr_x, we get         */
  /* something approximately equal to 2^MAX_PRECISION         */
  mpfr_init2(incr_x, 53);
  mpfr_set_ui(incr_x, MAX_EXPONENT + 1 - MIN_EXPONENT, GMP_RNDU);
  mpfr_div_ui(incr_x, incr_x, POINTS_X, GMP_RNDU);
  mpfr_exp2(incr_x, incr_x, GMP_RNDU);

  mpfr_clear(temp);
  mpfr_clear(temp2);

  /* Main loop */
  mpfr_init(x);
  prec = MIN_PRECISION;
  for(i=0; i<POINTS_PREC; i++) {
    mpfr_set_prec(x, prec);
    mpfr_set(x, x0, GMP_RNDU);
    for(j=0; j<POINTS_X; j++) {
      eq1 = timing_mpfr(x, prec);
      eq2 = timing_new(x, prec);

      best = (eq1<eq2)?1:2;

      fprintf(output, "%f\t", mpfr_get_d(x, GMP_RNDN));
      fprintf(output, "%lu\t", (unsigned long)prec);
      fprintf(output, "%f\t", eq1);
      fprintf(output, "%f\t", eq2);
      fprintf(output, "%d\n", best);

      mpfr_mul(x, x, incr_x, GMP_RNDU);
    }
    prec = (mp_prec_t)((double)prec * incr_prec + 1.);
    fprintf(output, "\n");
  }

  mpfr_clear(incr_x);
  mpfr_clear(x);
  mpfr_clear(x0);
  fclose(output);
  return 0;
}
